open Ast
open Printf

open Common

let rec string_of_expr modules e =
  match e with
    | Ident s -> s
    | Int i1 -> string_of_int i1
    | String s -> "\""^s^"\""
    | Quotation (s, _) -> s
    | Plus (e1 , e2) ->
        (string_of_expr modules e1)^" + "^(string_of_expr modules e2)
    | App (e1 , e2) ->
        let list_of_app e =
          let rec aux e acc =
            match e with
              | App (e1, e2) -> aux e1 (e2::acc)
              | _ -> e::acc
          in
            aux e []
        in
        let exprs = list_of_app e in
          (match exprs with
            | [] -> failwith "No."
            | x::q ->
                (string_of_expr modules x)^(string_of_list (string_of_expr modules) "(" ")" ", " q)
          )
    | Value (path, name) ->
        let rec aux i l =
          match l with
            | [] -> failwith "Unbound module"
            | (m, v)::q ->
                let rec aux2 j l =
                  match l with
                    | [] -> failwith "Unbound value in module"
                    | v::q ->
                        if v = name then
                          j
                        else
                          aux2 (j+1) q
                in
                  if m = path then
                    i, aux2 0 v
                  else
                    aux (i+1) q
        in
        let i, j = aux 0 modules in
          sprintf "__nol_module_%i_values_%i" i j

let expr modules file e =
  fprintf file "%s" (string_of_expr modules e)

let decls modules file decls =
  let decl (name, args, e, t) =
    match args with
      | [] ->
          fprintf file "int %s = " name;
          expr modules file e;
          print_endline ";";
          ()
      | l ->
          let args = string_of_list (fun x -> sprintf "int %s" x) "" "" ", " l in
          fprintf file "int %s(%s){\n" name args;
          expr modules file e;
          fprintf file ";\n}\n\n";
          ()
  in
    List.iter decl decls;
    ()

let quotations file q =
  List.iter (fun (s, _) -> fprintf file "%s\n" s) q;
  fprintf file "\n";
  ()

let stms modules file exprs =
  List.iter (fun x -> expr modules file x; fprintf file ";\n") exprs

let headers file =
  fprintf file "#include <dlfcn.h>\n";
  fprintf file "\n";
  ()

let modules_decls cfile modules =
  let rec aux i l =
    match l with
      | [] -> ()
      | (_, values)::q ->
          let rec aux2 j l =
            match l with
              | [] -> ()
              | _::q -> 
                  fprintf cfile "static void (*__nol_module_%i_values_%i)(char*);\n" i j;
                  aux2 (j+1) q;
                  ()
          in
            fprintf cfile "static void* __nol_module_%i;\n" i;
            aux2 0 values;
            aux (i+1) q
  in
    aux 0 modules;
    fprintf cfile "\n";
    ()

let modules_load file modules =
  let rec aux i l =
    match l with
      | [] -> ()
      | (path, values)::q ->
          let rec aux2 j l =
            match l with
              | [] -> ()
              | x::q -> 
                  fprintf file ("__nol_module_%i_values_%i = dlsym(__nol_module_%i, \"%s\");\n") i j i x;
                  aux2 (j+1) q;
                  ()
          in
          let path = string_of_list (fun x -> x) "" "" "/" path in
            fprintf file "__nol_module_%i = dlopen(\"./%s.so\", RTLD_NOW);\n" i path;
            aux2 0 values;
            aux (i+1) q
  in
    aux 0 modules;
    fprintf file "\n";
    ()

let modules_close file modules =
  let rec aux i l =
    match l with
      | [] -> ()
      | _::q ->
          fprintf file "dlclose(__nol_module_%i);\n" i;
          aux (i+1) q
  in
    aux 0 modules;
    fprintf file "\n";
    ()

let init file modules =
  fprintf file "int __nol_init() {\n";
  modules_load file modules;
  fprintf file "}\n\n";
  ()

let fini file modules =
  fprintf file "int __nol_fini() {\n";
  modules_close file modules;
  fprintf file "}\n\n";
  ()


let main modules file exprs =
  fprintf file "int main(int argc, char* argv[]) {\n";
  fprintf file "__nol_init();\n";
  stms modules file exprs;
  fprintf file "__nol_fini();\n";
  fprintf file "return 0;\n";
  fprintf file "}\n";
  ()

let compile ismodule target name =
  let options =
    if ismodule then
      "-shared -fPIC"
    else
      ""
  in
  let gcc = "gcc "^options^" "^name^" -o "^target^" -ldl -lgc" in
    print_endline (name^" -> "^target);
    print_endline gcc;
    ignore (Unix.system gcc);
    ()

let run target =
  ignore (Unix.execv target [||])

let print ismodule src_name ast =
  let (cname, cfile) = Filename.open_temp_file "nol_" ".c" in
  let target =
    match src_name with
      | None -> "a.out"
      | Some f -> Filename.chop_extension f
  in
  let target =
    if ismodule then
      target^".so"
    else
      target
  in
  let q = stm2quot (List.filter (function Quotation_stm _ -> true | _ -> false) ast) in
  let e = stm2expr (List.filter (function Expr _ -> true | _ -> false) ast) in
  let d = stm2decl (List.filter (function Decl _ -> true | _ -> false) ast) in
  let m = Modules.find ast in
    headers cfile;
    modules_decls cfile m;
    quotations cfile q;
    decls m cfile d;
    init cfile m;
    fini cfile m;
    if not ismodule then
      main m cfile e;
    close_out cfile;
    compile ismodule target cname;
    if not ismodule then
      run target
    else
      ()

